#include "lusol.h"

/*-------------------------------------------------*/
/*         non-level scheduling version            */
/*            16-thread L/U kernel                 */
/*            rows are sequential                  */
/*         each row is split to 16 pieces          */
/*-------------------------------------------------*/
__global__ 
void LUSol(int n, REAL *b, REAL *x, REAL *la, int *lja, 
           int *lia, REAL *ua, int *uja, int *uia) {
  int i,k,i1,i2;

  // thread lane in each half warp
  int lane = threadIdx.x;
  // shared memory for patial result, 24 == 16+8
  volatile __shared__ REAL r[24];
  
  /* Forward solve. Solve L*x = b */
  for (i=0; i<n; i++) {
    i1 = lia[i];
    i2 = lia[i+1];
    REAL sum = 0.0;

    for (k=i1+lane; k<i2; k+=HALFWARP)
      sum += la[k-1]*x[lja[k-1]-1];

    // parallel reduction
    r[threadIdx.x] = sum;
    r[threadIdx.x] = sum = sum + r[threadIdx.x+8];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+4];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+2];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+1];

    if (lane == 0)
      x[i] = b[i] - r[threadIdx.x];
  }
  
  /* Backward slove. Solve x = U^{-1}*x */
  for (i=n-1; i>=0; i--) {
    i1 = uia[i]+1;
    i2 = uia[i+1];

    REAL sum = 0.0;
    for (k=i1+lane; k<i2; k+=HALFWARP)
      sum += ua[k-1]*x[uja[k-1]-1];

    // parallel reduction
    r[threadIdx.x] = sum;
    r[threadIdx.x] = sum = sum + r[threadIdx.x+8];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+4];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+2];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+1];

    if (lane == 0) {
      REAL t = ua[i1-2];
      x[i] = t*(x[i]-r[threadIdx.x]);
    }
  }
}

/*---------------------------------------------------*
 *             d_x =  U^{-1}*L^{-1}*d_b
 *---------------------------------------------------*/
void luSolv(int n, int nnz, struct lu_t *d_lu, 
            REAL *d_x, REAL *d_b) {
  int i;
  double t1,t2;
  t1 = wall_timer();

  for (i=0; i<REPEAT; i++) {
    LUSol<<<1,16>>>(n, d_b, d_x, d_lu->l.a, d_lu->l.ja, 
                    d_lu->l.ia, d_lu->u.a, d_lu->u.ja, 
                    d_lu->u.ia);
  }
  // Barrier for GPU calls
  cudaThreadSynchronize();
  t2 = wall_timer() - t1;
  printf("[GPU] non-level-scheduling\n");
  printf("  time(s)=%.2f, Mflops=%.2f\n", t2,
         REPEAT*2*nnz/t2/1e6);
}

/*-------------------------------------------*/
/*     level scheduling kernel               */
/*-------------------------------------------*/
__global__
void L_SOL_LEV(REAL *b, REAL *x, REAL *la, 
               int *lja, int *lia,
               int *jlevL, int l1, int l2) {
  int i,k,jj;

  // num of half-warps
  int nhw = gridDim.x*BLOCKDIM/HALFWARP;
  // half warp id
  int hwid = (blockIdx.x*BLOCKDIM+threadIdx.x)/HALFWARP;
  // thread lane in each half warp
  int lane = threadIdx.x & (HALFWARP-1);
  // shared memory for patial result
  volatile __shared__ REAL r[BLOCKDIM+8];

  for (i=l1+hwid; i<l2; i+=nhw) {
    jj = jlevL[i-1]-1;
    int p1 = lia[jj];
    int q1 = lia[jj+1];

    REAL sum = 0.0;
    for (k=p1+lane; k<q1; k+=HALFWARP)
      sum += la[k-1]*x[lja[k-1]-1];

    // parallel reduction
    r[threadIdx.x] = sum;
    r[threadIdx.x] = sum = sum + r[threadIdx.x+8];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+4];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+2];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+1];

    if (lane == 0)
      x[jj] = b[jj] - r[threadIdx.x];
  }
}

/*----------------- x = U^{-1}*x */
__global__
void U_SOL_LEV(REAL *x, REAL *ua, int *uja, int *uia,
               int *jlevU, int l1, int l2) {
  int i,k,jj;

  // num of half-warps
  int nhw = gridDim.x*BLOCKDIM/HALFWARP;
  // half warp id
  int hwid = (blockIdx.x*BLOCKDIM+threadIdx.x)/HALFWARP;
  // thread lane in each half warp
  int lane = threadIdx.x & (HALFWARP-1);
  // shared memory for patial result
  volatile __shared__ REAL r[BLOCKDIM+8];

  for (i=l1+hwid; i<l2; i+=nhw) {
    jj = jlevU[i-1]-1;
    int p1 = uia[jj];
    int q1 = uia[jj+1];

    REAL sum = 0.0;
    for (k=p1+1+lane; k<q1; k+=HALFWARP)
      sum += ua[k-1]*x[uja[k-1]-1];

    // parallel reduction
    r[threadIdx.x] = sum;
    r[threadIdx.x] = sum = sum + r[threadIdx.x+8];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+4];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+2];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+1];

    if (lane == 0) {
      REAL t = ua[p1-1];
      x[jj] = t*(x[jj]-r[threadIdx.x]);
    }
  }
}

//--------------------------------------------------------
void luSolvLev(int n, int nnz, struct lu_t *d_lu, 
               struct level_t *h_lev,
               struct level_t *d_lev, REAL *d_x, REAL *d_b)
{
  int i,j;
  double t1, t2;

  t1 = wall_timer();
  for (j=0; j<REPEAT; j++) {
  // L-solve
  for (i=0; i<h_lev->nlevL; i++) {
    int l1 = h_lev->ilevL[i];
    int l2 = h_lev->ilevL[i+1];
    int l_size = l2 - l1;
    int nthreads = min(l_size*HALFWARP, MAXTHREADS);
    int gDim = (nthreads+BLOCKDIM-1)/BLOCKDIM;
    int bDim = BLOCKDIM;
    L_SOL_LEV<<<gDim, bDim>>>(d_b, d_x, d_lu->l.a, d_lu->l.ja,
                                  d_lu->l.ia, d_lev->jlevL, l1, l2);
  }
  // U-solve
  for (i=0; i<h_lev->nlevU; i++) {
    int l1 = h_lev->ilevU[i];
    int l2 = h_lev->ilevU[i+1];
    int l_size = l2 - l1;
    int nthreads = min(l_size*HALFWARP, MAXTHREADS);
    int gDim = (nthreads+BLOCKDIM-1)/BLOCKDIM;
    int bDim = BLOCKDIM;
    U_SOL_LEV<<<gDim, bDim>>>(d_x, d_lu->u.a, d_lu->u.ja,
                                  d_lu->u.ia, d_lev->jlevU, l1,l2);
  }
  }
  //Barrier for GPU calls
  cudaThreadSynchronize();
  t2 = wall_timer() - t1;
  printf("[GPU] level-scheduling, #lev in L %d, #lev in U %d\n",
  h_lev->nlevL, h_lev->nlevU);
  printf("  time(s)=%.2f, Mflops=%.2f\n", 
  t2, REPEAT*2*nnz/t2/1e6);
}

/*----------------------------------------------------------------*/
void luSolCPU(int n, int nnz, REAL *b, REAL *x, REAL *la, int *lja,
              int *lia, REAL *ua, int *uja, int *uia)
{
  int i,k,i1,i2,ii;
  double t1 = wall_timer();
  
  for (ii=0; ii<REPEAT; ii++) {
  /* Forward solve. Solve L*x = b */
  for (i=0; i<n; i++) {
    x[i] = b[i];
    i1 = lia[i];
    i2 = lia[i+1];
    for (k=i1; k<i2; k++)
      x[i] -= la[k-1]*x[lja[k-1]-1];
  }
  
  /* Backward slove. Solve x = U^{-1}*x */
  for (i=n-1; i>=0; i--) {
    REAL t = ua[uia[i]-1];
    i1 = uia[i]+1;
    i2 = uia[i+1];
    for (k=i1; k<i2; k++)
      x[i] -= ua[k-1]*x[uja[k-1]-1];

    x[i] = t*x[i];
  }
  }
  double t2 = wall_timer() - t1;
  printf("[CPU]\n");
  printf("  time(s)=%.2f, Mflops=%.2f\n", t2, 
          REPEAT*2*nnz/t2/1e6);
}
